/************************************************************************/
/*                                                                      */
/*    vspline - a set of generic tools for creation and evaluation      */
/*              of uniform b-splines                                    */
/*                                                                      */
/*            Copyright 2015 - 2017 by Kay F. Jahnke                    */
/*                                                                      */
/*    The git repository for this software is at                        */
/*                                                                      */
/*    https://bitbucket.org/kfj/vspline                                 */
/*                                                                      */
/*    Please direct questions, bug reports, and contributions to        */
/*                                                                      */
/*    kfjahnke+vspline@gmail.com                                        */
/*                                                                      */
/*    Permission is hereby granted, free of charge, to any person       */
/*    obtaining a copy of this software and associated documentation    */
/*    files (the "Software"), to deal in the Software without           */
/*    restriction, including without limitation the rights to use,      */
/*    copy, modify, merge, publish, distribute, sublicense, and/or      */
/*    sell copies of the Software, and to permit persons to whom the    */
/*    Software is furnished to do so, subject to the following          */
/*    conditions:                                                       */
/*                                                                      */
/*    The above copyright notice and this permission notice shall be    */
/*    included in all copies or substantial portions of the             */
/*    Software.                                                         */
/*                                                                      */
/*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND    */
/*    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES   */
/*    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND          */
/*    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT       */
/*    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,      */
/*    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      */
/*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR     */
/*    OTHER DEALINGS IN THE SOFTWARE.                                   */
/*                                                                      */
/************************************************************************/

/*! \file eval.h

    \brief code to evaluate uniform b-splines

    This body of code contains class evaluator and auxilliary classes which are
    needed for it's smooth operation.

    The evaluation is a reasonably straightforward process: A subset of the coefficient
    array, containing coefficients 'near' the point of interest, is picked out, and
    a weighted summation over this subset produces the result of the evaluation.
    The complex bit is to have the right coefficients in the first place
    (this is what prefiltering does), and to use the appropriate weights on
    the coefficient window. For b-splines, there is an efficient method to
    calculate the weights by means of a matrix multiplication, which is easily
    extended to handle b-spline derivatives as well. Since this code lends itself
    to a generic implementation, and it can be parametrized by the spline's order,
    and since the method performs well, I use it here in preference to the code
    which Thevenaz uses (which is, for the orders of splines it encompasses, the matrix
    multiplication written out with a few optimizations, like omitting multiplications
    with zero, and slightly more concise calculation of powers)
    
    Evaluation of a b-spline is, compared to prefiltering, more computationally intensive
    and less memory-bound, so the profit from vectorization, especially for float data,
    is more pronounced here than for the prefiltering. On my system, I found single-precision
    operation was about three to four times as fast as unvectorized code (AVX2).
    
    The central class of this file is class evaluator. evaluator objects are set up to
    provide evaluation of a specific b-spline. Once they are set up they don't change and
    effectively become pure functors with several overloaded evaluation methods for different
    constellations of parameters. The evaluation methods typically take their arguments per
    reference. The details of the evaluation variants, together with explanations of
    specializations used for extra speed, can be found with the individual evaluation
    routines.
    
    What do I mean by the term 'pure' functor? It's a concept from functional programming.
    It means that calling the functor will not have any effect on the functor itself - it
    can't change once it has been constructed. This has several nice effects: it can
    potentially be optimized very well, it is thread-safe, and it will play well with
    functional programming concepts - and it's conceptionally appealing.
    
    Code using class evaluator will probably use it at some core place where it is
    part of some processing pipeline. An example would be an image processing program:
    one might have some outer loop generating arguments (typically simdized types)
    which are processed one after the other to yield a result. The processing will
    typically have several stages, like coordinate generation and transformations,
    then use class evaluator to pick an interpolated intermediate result, which is
    further processed by, say, colour or data type manipulations before finally
    being stored in some target container. The whole processing pipeline can be
    coded to become a single functor, with one of class evaluator's eval-type
    routines embedded somewhere in the middle, and all that's left is code to
    efficiently handle the source and destination to provide arguments to the
    pipeline - like the code in remap.h. And since the code in remap.h is made to
    provide the data feeding and storing, the only coding needed is for the pipeline,
    which is where the 'interesting' stuff happens.
*/

#ifndef VSPLINE_EVAL_H
#define VSPLINE_EVAL_H

#include "bspline.h"
#include "unary_functor.h"

namespace vspline {

using namespace std ;
using namespace vigra ;

// is_singular tests if a type is either a plain fundamental or a Vc SIMD type.
// The second possibility is only considered if Vc is used at all.
// This test serves to differentiate between nD values like vigra TinyVectors
// which fail the test and singular values, which pass. Note that this test
// fails vigra::TinyVector < T , 1 > even though one might consider it 'singular'.

template < class T >
using is_singular = typename
  std::conditional
  <    std::is_fundamental < T > :: value
#ifdef USE_VC
    || Vc::is_simd_vector < T > :: value
#endif
    ,
    std::true_type ,
    std::false_type
  > :: type ;

// next we have coordinate splitting functions. For odd splines, coordinates
// are split into an integral part and a remainder between 0 and 1, which is
// used for weight generation.
// we have two variants for odd_split and the dispatch below

template < typename ic_t , typename rc_t , int vsize = 1 >
void odd_split ( rc_t v , ic_t& iv , rc_t& fv , std::true_type )
{
  rc_t fl_i = std::floor ( v ) ;
  fv = v - fl_i ;
  iv = ic_t ( fl_i )  ;
}

template < typename ic_t , typename rc_t , int vsize = 1 >
void odd_split ( rc_t v , ic_t& iv , rc_t& fv , std::false_type )
{
  for ( int d = 0 ; d < vigra::ExpandElementResult < rc_t > :: size ; d++ )
    odd_split ( v[d] , iv[d] , fv[d] , std::true_type() ) ;
}

template < typename ic_t , typename rc_t , int vsize = 1 >
void odd_split ( rc_t v , ic_t& iv , rc_t& fv )
{
  odd_split ( v , iv , fv , is_singular<rc_t>() ) ;
}

// for even splines, the integral part is obtained by rounding. when the
// result of rounding is subtracted from the original coordinate, a value
// between -0.5 and 0.5 is obtained which is used for weight generation.
// we have two variants for even_split and the dispatch below

template < typename ic_t , typename rc_t , int vsize = 1 >
void even_split ( rc_t v , ic_t& iv , rc_t& fv , std::true_type )
{
  rc_t fl_i = std::round ( v ) ;
  fv = v - fl_i ;
  iv = ic_t ( fl_i ) ;
}

template < typename ic_t , typename rc_t , int vsize = 1 >
void even_split ( rc_t v , ic_t& iv , rc_t& fv , std::false_type )
{
  for ( int d = 0 ; d < vigra::ExpandElementResult < rc_t > :: size ; d++ )
    even_split ( v[d] , iv[d] , fv[d] , std::true_type() ) ;
}

template < typename ic_t , typename rc_t , int vsize = 1 >
void even_split ( rc_t v , ic_t& iv , rc_t& fv )
{
  even_split ( v , iv , fv , is_singular<rc_t>() ) ;
}

// TODO describe the use of the 'weight matrix' rather than technical details

/// The routine 'calculate_weight_matrix' originates from vigra. I took the original
/// routine BSplineBase<spline_order, T>::calculateWeightMatrix() from vigra and changed it
/// in several ways:
///
/// - the spline degree is now a runtime parameter, not a template argument
/// - the derivative degree is passed in as an additional parameter, directly
///   yielding the appropriate weight matrix needed to calculate a b-spline's derivative
///   with only a slight modification to the original code
/// - the code uses my modified bspline basis function which takes the degree as a
///   run time parameter instead of a template argument and works with integral
///   operands and precalculated values, which makes it very fast, even for high
///   spline degrees. bspline_basis() is in basis.h.

template < class target_type >
MultiArray < 2 , target_type > calculate_weight_matrix ( int degree , int derivative )
{
  const int order = degree + 1 ;
  
  if ( derivative >= order ) // guard against impossible parameters
    return MultiArray < 2 , target_type >() ;

  // allocate space for the weight matrix
  MultiArray < 2 , target_type > res = MultiArray < 2 , target_type > ( order , order - derivative ) ;
  
  long double faculty = 1.0 ;
  
  for ( int row = 0 ; row < order - derivative ; row++ )
  {
    if ( row > 1 )
      faculty *= row ;

    int x = degree / 2 ; // (note: integer division)

    // we store to a MultiArray, which is row-major, so storing as we do
    // places the results in memory in the precise order in which we want to
    // use them in the weight calculation.
    // note how we pass x to bspline_basis() as an integer. This way, we pick
    // a very efficient version of the basis function which only evaluates at
    // whole numbers. This basis function version does hardly any calculations
    // but instead relies on precalculated values. see bspline_basis in prefilter.h
    // note: with large degrees (20+), this still takes a fair amount of time, but
    // rather seconds than minutes with the standard routine.
    
    for ( int column = 0 ; column < order ; ++column , --x )
      res ( column , row ) = bspline_basis<long double> ( x , degree , row + derivative ) / faculty;
  }

  return res;
}

/// while we deal with B-splines here, there is no need to limit the evaluator
/// code only to B-spline basis functions. The logic is the same for any type of evaluation
/// which functions like a separable convolution of an equilateral subarray of the coefficient
/// array, and the only thing specific to b-splines is the weight generation.
///
/// So I coded the general case, which can use any weight generation function. Coding this
/// introduces a certain degree of complexity, which I feel is justified for the flexibility
/// gained. The complexity is mainly due to the fact that, while
/// we can write a simple (templated) function to generate weights (as above), we can't pass
/// such a template as an object to a function. Instead we use an abstract base class for
/// the weight functor and inherit from it for specific weight generation methods.
///
/// I made some investigations towards coding evaluation of splines with different orders
/// along the different axes, but this introduced too much extra complexity for my taste and
/// took me too far away from simply providing efficient code for b-splines, so I abandoned
/// the attempts. Therefore the weight functors for a specific spline all have to have a common
/// spline_order and generate spline_order weights. The only way to force lesser order weight
/// functors into this scheme is to set some of the weights to zero. Weight functors
/// of higher spline_order than the spline can't be accomodated, if that should be necessary,
/// the spline_order of the entire spline has to be raised.
///
/// Note the use of 'delta' in the functions below. this is due to the fact that these functors
/// are called with the fractional part of a real valued coordinate.
///
/// first we define a base class for (multi-)functors calculating weights.
/// this base class can accomodate weight calculation with any weight generation
/// function using the same signature. It is not specific to b-splines.
/// We access the weight functors via a pointer to this base class in the code below.

template < typename ele_type , // elementary type of value_type
           typename rc_type ,  // type of real-valued coordinate
           int vsize = 1 >
struct weight_functor_base
{
  // we define two pure virtual overloads for operator(), one for unvectorized
  // and one for vectorized operation. In case the scope of evaluation is extended
  // to other types of values, we'll have to add the corresponding signatures here.
  
  virtual void operator() ( ele_type* result , const rc_type& delta ) = 0 ;
  
#ifdef USE_VC

  typedef typename vector_traits < ele_type , vsize > :: ele_v ele_v ;
  typedef typename vector_traits < rc_type , vsize > :: ele_v rc_v ;

  virtual void operator() ( ele_v* result , const rc_v& delta ) = 0 ;

#endif
} ;

/// this functor calculates weights for a b-spline or it's derivatives.
/// with d == 0, the weights are calculated for plain evaluation.
/// Initially I implemented weight_matrix as a static member, hoping the code
/// would perform better, but I could not detect significant benefits. Letting
/// the constructor choose the derivative gives more flexibility and less type
/// proliferation.

template < typename target_type ,   // type for weights (may be a simdized type)
           typename ele_type ,      // elementary type of value_type
           typename delta_type >    // type for deltas (may be a simdized type)

struct bspline_derivative_weights
{
  typedef typename MultiArray < 2 , ele_type > :: iterator wm_iter ;

  // TODO I would make this const, but in vigra, the iterator obtained by calling begin()
  // on a const array is protected (multi_ierator.hxx, 438) why is this so?

  MultiArray < 2 , ele_type > weight_matrix ;
  const int degree ;
  const int derivative ;
  const int columns ;
  wm_iter wm_begin ;
  wm_iter wm_end ;

  bspline_derivative_weights ( int _degree , int _derivative = 0 )
  : weight_matrix ( calculate_weight_matrix < ele_type > ( _degree , _derivative ) ) ,
    degree ( _degree ) ,
    derivative ( _derivative ) ,
    columns ( _degree + 1 )
  { 
    wm_begin = weight_matrix.begin() ;
    wm_end = weight_matrix.end() ;
  } ;
  
  void operator() ( target_type* result , const delta_type & delta )
  {
    target_type power ( delta ) ;
    wm_iter factor_it = wm_begin ;
    const wm_iter factor_end = wm_end ;

    // the result is initialized with the first row of the 'weight matrix'.
    // We save ourselves multiplying it with delta^0.
 
    for ( int c = 0 ; c < columns ; c++ )
    {
      result[c] = *factor_it ;
      ++factor_it ;
    }
    
    if ( degree )
    {
      for ( ; ; )
      {
        for ( int c = 0 ; c < columns ; c++ )
        {
          result[c] += power * *factor_it ;
          ++factor_it ;
        }
        if ( factor_it == factor_end ) // avoid multiplication if exhausted, break now
          break ;
        power *= target_type ( delta ) ; // otherwise produce next power(s) of delta(s)
      }
    }
  }
} ;

/// we derive from the weight functor base class to obtain a (multi-) functor
/// specifically for (derivatives of) a b-spline :

template < typename ele_type , typename rc_type , int vsize = 1 >
struct bspline_derivative_weight_functor
: public weight_functor_base < ele_type , rc_type , vsize >
{
  typedef weight_functor_base < ele_type , rc_type , vsize > base_class ;

  // set up the fully specialized functors to which operator() delegates:

  bspline_derivative_weights < ele_type , ele_type , rc_type >  weights ;

#ifdef USE_VC
  using typename base_class::ele_v ;
  using typename base_class::rc_v ;
  
  bspline_derivative_weights < ele_v , ele_type , rc_v >  weights_v ; 
#endif

  bspline_derivative_weight_functor ( int degree , int d = 0 )
  : weights ( degree , d )
#ifdef USE_VC
  , weights_v ( degree , d )
#endif
  {
  }
  
  // handle the weight calculation by delegation to the functors set up at construction
  
  virtual void operator() ( ele_type* result , const rc_type& delta )
  {
    weights ( result , delta ) ;
  }

#ifdef USE_VC
  virtual void operator() ( ele_v* result , const rc_v& delta )
  {
    weights_v ( result , delta ) ;
  }
#endif
} ;

// not a very useful weight function, but it suits to prove the concept of plug-in
// weight functions works as intended. Instead of calculating the weights, as in the functor
// above, this one simply returns equal weights. The result is that, no matter what delta
// is passed in, the weights are the same and average over the coefficients to which they
// are applied.
// Note here that one important property of the weights is that they constitute
// a partition of unity. Both the (derivative) b-spline weights and this simple
// weight functor share this property.
// currently unused, but left in for demonstration purposes

/*

template < typename rc_type >
struct average_weight_functor
: public weight_functor_base < rc_type >
{
  typedef weight_functor_base < rc_type > base_class ;
  using typename base_class::rc_type_v ;

  const rc_type weight ;
  const int order ;
  
  average_weight_functor ( int degree )
  : weight ( rc_type ( 1.0 ) / rc_type ( degree + 1 ) ) ,
    order ( degree + 1 )
  { } ;
  
  virtual void operator() ( rc_type* result , const rc_type& delta )
  {
    for ( int e = 0 ; e < order ; e++ )
      result[e] = weight ;
  }
  
  virtual void operator() ( rc_type_v* result , const rc_type_v& delta )
  {
    for ( int e = 0 ; e < order ; e++ )
      result[e] = weight ;
  }  
} ;

*/

/// class evaluator encodes evaluation of a B-spline. The evaluation relies on 'braced'
/// coefficients, as they are normally provided by a vspline::bspline object (the exception
/// being bspline objects created with UNBRACED or MANUAL strategy). While the most
/// general constructor will accept a MultiArrayView to coefficients (including the necessary
/// 'brace'), this will rarely be used, and an evaluator will be constructed from a bspline
/// object. In the most trivial case there are only two things which need to be done:
///
/// The specific type of evaluator has to be established by providing the relevant template
/// arguments. Here, we need two types: the 'coordinate type' and the 'value type'.
///
/// - The coordinate type is encoded as a vigra::TinyVector of some real data type - if you're
/// doing image processing, the typical type would be a vigra::TinyVector < float , 2 >.
///
/// - The value type has to be either an elementary real data type such as 'float' or 'double',
/// or a vigra::TinyVector of such an elementary type. Other data types which can be handled
/// by vigra's ExpandElementResult mechanism should also work. When processing colour images,
/// your value type would typically be a vigra::TinyVector < float , 3 >.
///
/// Additionally, the width of SIMD vectors to be used has to be established. This is not
/// mandatory - if omitted, a default value is picked.
///
/// With the evaluator's type established, an evaluator of this type can be constructed by
/// passing a vspline::bspline object to the constructor. Naturally, the bspline object has
/// to contain data of the same value type, and the spline has to have the same number of
/// dimensions as the coordinate type.
/// 
/// I have already hinted at the evaluation process used, but here it is again in a nutshell:
/// The coordinate at which the spline is to be evaluated is split into it's integral part
/// and a remaining fraction. The integral part defines the location where a window from the
/// coefficient array is taken, and the fractional part defines the weights to use in calculating
/// a weighted sum over this window. This weighted sum represents the result of the evaluation.
/// Coordinate splitting is done with the method split(), which picks the appropriate variant
/// (different code is needed for odd and even splines)
///
/// The generation of the weights to be applied to the window
/// of coefficients is performed by employing the weight functors above. What's left to do is
/// to bring all the components together, which happens in class evaluator. The workhorse
/// code in the subclasses _eval and _v_eval takes care of performing the necessary operations
/// recursively over the dimensions of the spline.
///
/// a vspline::evaluator is technically a vspline::unary_functor. This way, it can be directly
/// used by constructs like vspline::chain and has a consistent interface which allows code
/// using evaluators to query it's specifics. The ectual evaluation capabilites (eval() routines)
/// are incorporated into the vspline::unary_functor as a policy - class evaluator_policy
/// below. The final type, vspline::evaluator, is fixed by a using declaration right at the
/// end of this file.
///
/// The code in evaluator_policy begins with a sizeable constructor which sets up as much
/// as possible to support the evaluation code to do it's job as fast as possible. Next follows
/// the unvectorized code, finally the vectorized code.
///
/// There is a variety of overloads of class evaluator_policy's eval() method available, which
/// become methods of vspline::evaluator via the policy incorporation. Some of these overloads
/// go beyond the 'usual' eval routines, like evaluation with known weights, sidestepping the
/// weight generation from coordinates.
///
/// The evaluation strategy is to have all dependencies of the evaluation except for the actual
/// coordinates taken care of by the constructor - and immutable for the evaluator's lifetime.
/// The resulting object has no state which is modified after construction, making it thread-safe.
/// It also constitutes a 'pure' functor in a functional-programming sense, because it has
/// no mutable state and no side-effects, as can be seen by the fact that the eval methods
/// are all marked const.
///
/// The eval() overloads form a hierarchy, as evaluation progresses from accepting unsplit real
/// coordinates to split coordinates and finally offsets and weights. This allows calling code to
/// handle parts of the delegation hierarchy itself, only using an evaluator at a specific level.
///
/// By providing the evaluation in this way, it becomes easy for calling code to integrate
/// the evaluation into more complex functors. Consider, for example, code
/// which generates coordinates with a functor, then evaluates a b-spline at these coordinates,
/// and finally subjects the resultant values to some postprocessing. All these processing
/// steps can be bound into a single functor, and the calling code can be reduced to polling
/// this functor until it has obtained the desired number of output values.
///
/// While the 'unspecialized' evaluator will try and do 'the right thing' by using general
/// purpose code fit for all eventualities, for time-critical operation there is a
/// specialization which can be used to make the code faster:
///
/// - template argument 'specialize' can be set to 0 to forcibly use (more efficient) nearest
/// neighbour interpolation, which has the same effect as simply running with degree 0 but avoids
/// code which isn't needed for nearest neighbour interpolation (like the application of weights,
/// which is futile under the circumstances, the weight always being 1.0).
/// specialize can also be set to 1 for explicit n-linear interpolation. Any other value will
/// result in the general-purpose code being used. Note that the specialization is passed in
/// as a type, so instead of accepting a straight 'int' template argument,
/// std::integral_constant<int,SPEC> is required, This is for technical reasons:
/// vspline::unary_functor routes trailing template arguments to it's evaluation policy,
/// which is coded via a variadic template argument list. Such a list cannot have a mixture
/// of type and non-type template arguments, so I've gone for a type template argument list.
///
/// Note that, contrary to my initial implementation, all forms of coordinate mapping were
/// removed from class evaluator. The 'mapping' which is left is, more aptly, called
/// 'splitting', since the numeric value of the incoming coordinate is never modified.
/// Folding arbitrary coordinates into the spline's defined range now has to be done
/// externally, typically by wrapping class evaluator together with some coordinate
/// modification code into a combined vspline::unary_functor. map.h provides code for
/// common mappings, see there. Arbitrary manipulation of incoming and outgoing data
/// can be realized be using vspline::chain in unary_functor.h, see there.
///
/// Note how the default number of vector elements is fixed by picking the number of ele_type
/// which vspline::vector_traits considers appropriate. There should rarely be a need to
/// choose a different number of vector elements: evaluation will often be the most
/// computationally intensive part of a processing chain, and therefore this choice is
/// sensible. But it's not mandatory.

// TODO: we inherit from uf_types to use the standard evaluation
// type system, but I'd like to incorporate that into unary_functor.
// Can I have these types in some other way? I'd like to get rid
// of uf_types altogether, and this is the only other place where
// it's used.

// TODO: evaluator_policy uses a great many typedefs which is a bit over the top, even though
// it was helpful in the development process.

template < typename _coordinate_type , // nD real coordinate
           typename _value_type ,      // type of coefficient/result
#ifdef USE_VC
           // nr. of vector elements
           int _vsize = vspline::vector_traits < _value_type > :: size ,
#else
           int _vsize = 1 ,
#endif
           // specialize for degree 0 or 1 spline
           typename specialize = std::integral_constant<int,-1>
         >
class evaluator_policy
: public uf_types < _coordinate_type , _value_type , _vsize >
{
public:

  typedef _value_type value_type ; // == base_type::out_type

  // we want to access facilites of the base class (vspline::uf_types<...>)
  // so we use a typedef for the base class.

  typedef uf_types < _coordinate_type , _value_type , _vsize > base_type ;

  // pull in standard evaluation type system with this macro:

  using_unary_functor_types ( base_type ) ;
  
  // relying on base type to provide these:
  
  typedef in_ele_type rc_type ;    // elementary type of a coordinate
  typedef out_ele_type ele_type ;  // elementary type of value_type

  // Initially I was using a template argument for this flag, but it turned out
  // that using a const bool set at construction time performed just as well.
  // Since this makes using class evaluator easier, I have chosen to go this way.

  const bool even_spline_degree ;   // flag containing the 'evenness' of the spline

  enum { dimension = dim_in }  ;
  enum { level = dim_in - 1 }  ;
  enum { channels = dim_out } ;

  // types for nD integral indices, these are not in base_type
  
  typedef int ic_type ;            // we're only using int for indices
  
  typedef vigra::TinyVector < ic_type , dimension > nd_ic_type ;
  
  // we don't use:  typedef in_type nd_rc_type ;
  // because if the spline is 1D, in_type comes out as, say, a plain float,
  // while the code in class evaluator expects TinyVector<float,1>. This produces
  // no inconvenience, since there are specializations for 1D splines.

  typedef vigra::TinyVector < rc_type , dimension > nd_rc_type ;

  /// view_type is used for a view to the coefficient array

  typedef MultiArrayView < dimension , value_type >                view_type ;

  /// type used for nD array coordinates, array shapes

  typedef typename view_type::difference_type                      shape_type ;

  typedef vigra::TinyVector < int , dimension >                    derivative_spec_type ;

  typedef typename MultiArrayView < 1 , ic_type > :: const_iterator offset_iterator ;
  
  typedef vigra::MultiCoordinateIterator<dimension>                 nd_offset_iterator ;
  
  typedef weight_functor_base < ele_type , rc_type , vsize > weight_functor_base_type ;

  /// in the context of b-spline calculation, this is the weight generating
  /// functor which will be used:

  typedef bspline_derivative_weight_functor < ele_type , rc_type , vsize >
    bspline_weight_functor_type ;
  
  // to try out gaussian weights, one might instead use
  // typedef gaussian_weight_functor < ele_type > bspline_weight_functor_type ;
  
  /// we need one functor per dimension:
    
  typedef vigra::TinyVector < weight_functor_base_type* , dimension > nd_weight_functor ;
  
  // while in the context of B-splines the weight functors are, of course, functors which
  // calculate the weights via the B-spline basis functions, the formulation we use here
  // allows us to use any set of functors that satisfy the argument type. With this
  // flexible approach, trying out other basis functions becomes simple: just write the
  // functor and pass it in, it doesn't have to have anything to do with B-splines at all.
  // Another way to employ a different weight generation method is by passing different
  // weights into operator() below.
  // By default class evaluator will use b-spline weights.
  
private:
  
  nd_weight_functor fweight ;        ///< set of pointers to weight functors, one per dimension
  const value_type * const cf_base ; ///< coefficient base adress in terms of value_type
  const ele_type * const cf_ebase ;  ///< coefficient base adress in terms of ele_type
  const shape_type cf_shape ;           ///< shape in terms of value_type
  const shape_type cf_eshape ;          ///< shape in terms of ele_type
  const shape_type cf_stride ;          ///< strides in terms of value_type
  const shape_type cf_estride ;         ///< strides in terms of expanded value_type
  MultiArray < 1 , ic_type > offsets ;  ///< offsets in terms of value_type
  MultiArray < 1 , ic_type > eoffsets ; ///< offsets in terms of ele_type, for vc op
  bspline_weight_functor_type wfd0 ;    ///< default weight functor: underived bspline
  const int spline_degree ;
  const int spline_order ;
  const int window_size ;

public:

  const int & get_order() const
  {
    return spline_order ;
  }

  const int & get_degree() const
  {
    return spline_degree ;
  }

  const shape_type & get_stride() const
  {
    return cf_stride ;
  }

  /// this constructor is the most flexible variant and will ultimately be called by all other
  /// constructor overloads. This constructor will not usually be called directly - rather
  /// use the overloads taking vspline::bspline objects. The constructor takes three arguments:
  /// - a vigra::MultiArrayView to (braced) coefficients
  /// - the degree of the spline (3 == cubic spline)
  /// - specification of the desired derivative of the spline, defaults to 0 (plain evaluation).
  
  // TODO now that evaluator_policy doesn't hold a MultiArrayView anymore, it's easy to
  // provide a constructor which directly takes address, shape and stride of the coefficient
  // array to approach an interface which does not use vigra types.

  evaluator_policy ( const view_type & _coefficients ,
              int _spline_degree ,
              derivative_spec_type _derivative )
  : cf_base ( _coefficients.data() ) ,
    cf_ebase ( (ele_type*)(_coefficients.data()) ) ,
    cf_shape ( _coefficients.shape() ) ,
    cf_eshape ( channels * _coefficients.shape() ) ,
    cf_stride ( _coefficients.stride() ) ,
    cf_estride ( channels * _coefficients.stride() ) ,
    spline_degree ( _spline_degree ) ,
    even_spline_degree ( ! ( _spline_degree & 1 ) ) ,
    spline_order ( _spline_degree + 1 ) ,
    wfd0 ( _spline_degree , 0 ) ,
    window_size ( std::pow ( _spline_degree + 1 , int(dimension) ) )
  {
    // initalize the weight functors. In this constructor, we use only bspline weight
    // functors, even though the evaluator can operate with all weight functors
    // filling in the right number of basis values given a delta. To make use of this
    // flexibility, one would derive from this class or write another constructor.
    // Note how we code so that the default case (plain evaluation with no derivatives)
    // results in use of only one single weight functor.

    for ( int d = 0 ; d < dimension ; d++ )
    {
      if ( _derivative[d] )
      {
        fweight[d] = new bspline_weight_functor_type ( _spline_degree , _derivative[d] ) ;
      }
      else
      {
        fweight[d] = &wfd0 ; // pick the default if derivative is 0
      }
    }
    
    // calculate the number of offsets needed and create the array to hold them
    // The evaluation forms a weighted sum of a window into the coeffcicent array.
    // The sequence of offsets we calculate here is the set of pointer differences
    // from the first element in that window to all elements in the window. It's
    // another way of coding this window, where all index calculations have already
    // been done beforehand rather than performing it during the traversal of the
    // window by means of stride/shape arithmetic. Coding the window in this fashion
    // also makes it easy to vectorize the code.
    
    offsets = MultiArray < 1 , ptrdiff_t > ( window_size ) ;
    eoffsets = MultiArray < 1 , ptrdiff_t > ( window_size ) ;
    
    // we fill the offset array in a simple fashion: we do a traversal of the window
    // and calculate the pointer difference for every element reached. We use the
    // same loop to code the corresponding offsets to elementary values (ele_type)
  
    auto sample = _coefficients.subarray ( shape_type() , shape_type(spline_order) ) ;
    auto base = sample.data() ;
    auto target = offsets.begin() ;
    auto component_target = eoffsets.begin() ;

    for ( auto &e : sample )
    {
      *target = &e - base ;
      *component_target = channels * *target ;
      ++target ;
      ++component_target ;
    }
  } ;

  /// simplified constructors from a bspline object
  /// when using the higher-level interface to vspline's facilities - via class bspline -
  /// the bspline object provides the coefficients. The derivative specification is passed
  /// in just as for the general constructor. Note that the derivative specification can be
  /// individually chosen for each axis.
  
  evaluator_policy ( const bspline < value_type , dimension > & bspl ,
                     derivative_spec_type _derivative = derivative_spec_type ( 0 ) )
  : evaluator_policy ( bspl.coeffs ,
                       bspl.spline_degree ,
                      _derivative )
  {
    if ( bspl.spline_degree > 1 && ! bspl.braced )
      throw not_supported ( "for spline degree > 1: evaluation needs braced coefficients" ) ;
  } ;

  /// obtain_weights calculates the weights to be applied to a section of the coefficients from
  /// the fractional parts of the split coordinates. What is calculated here is the evaluation
  /// of the spline's basis function at dx, dx+1 , dx+2..., but doing it naively is computationally
  /// expensive, as the evaluation of the spline's basis function at arbitrary values has to
  /// look at the value, find out the right interval, and then calculate the value with the
  /// appropriate function. But we always have to calculate the basis function for *all*
  /// intervals anyway, and the method used here performs this tasks efficiently using a
  /// vector/matrix multiplication.
  ///
  /// If the spline is more than 1-dimensional, we need a set of weights for every dimension.
  /// The weights are accessed with a 2D vigra MultiArrayView.
  ///
  /// Contrary to my initial implementation, I fill the 'workspace' in order ascending with
  /// the axis, so now weights for axis 0 are first etc.. This results in a slighly funny-looking
  /// initial call to _eval, but the confusion from reverse-filling the workspace was probably worse.

  // TODO: change (*(fweight[axis]))() to work on 'weight' rather than on a pointer?
  
  template < typename nd_rc_type , typename weight_type >
  void obtain_weights ( const MultiArrayView < 2 , weight_type > & weight ,
                        const nd_rc_type& c ) const
  {
    auto ci = c.cbegin() ;
    for ( int axis = 0 ; axis < dimension ; ++ci , ++axis )
      (*(fweight[axis])) ( weight.data() + axis * spline_order , *ci ) ;
  }

  /// obtain weight for a single axis

  template < typename rc_type , typename weight_type >
  void obtain_weights ( weight_type * p_weight ,
                        const int & axis ,
                        const rc_type& c ) const
  {
    (*(fweight[axis])) ( p_weight , c ) ;
  }

  /// split function. This function is used to split incoming real coordinates
  /// into an integral and a remainder part, which are used at the core of the
  /// evaluation. selection of even or odd splitting is done via the const bool
  /// flag 'even_spline_degree' My initial implementation had this flag as a
  /// template argument, but this way it's more flexible and there seems to
  /// be no runtime penalty. This method delegates to the free function templates
  /// even_split and odd_split, respectively, which are defined above class evaluator.

  template < class IT , class RT >
  void split ( const RT& input , IT& select , RT& tune ) const
  {
    if ( even_spline_degree )
      even_split < IT , RT , vsize > ( input , select , tune ) ;
    else
      odd_split < IT , RT , vsize > ( input , select , tune ) ;
  }

  /// _eval is the workhorse routine and implements the recursive arithmetic needed to
  /// evaluate the spline. First the weights for the current dimension are obtained
  /// from the weights object passed in. Once the weights are known, they are successively
  /// multiplied with the results of recursively calling _eval for the next
  /// lower dimension and the products are summed up to produce the return value.
  /// The scheme of using a recursive evaluation has several benefits:
  /// - it needs no explicit intermediate storage of partial sums (uses stack instead)
  /// - it makes the process dimension-agnostic in an elegant way
  /// - therefore, the code is also thread-safe
  /// _eval works with a base pointer and an iterator over offsets, just like the vectorized version.
  /// note that this routine is used for operation on braced splines, with the sequence of offsets to be
  /// visited fixed at the evaluator's construction. But in this non-vectorized routine, passing in a
  /// different sequence of offsets for evaluation in an area where boundary conditions apply would be
  /// feasible (akin to Thevenaz' indexing), even though it would require a lot of new logic, since
  /// currently the bracing takes care of the boundary conditions.
  ///  
  /// I'd write a plain function template and partially specialize it for 'level', but that's
  /// not allowed, so I use a functor instead.
  
  template < int level , class dtype >
  struct _eval
  {
    dtype operator() ( const dtype* const & pdata ,
                       offset_iterator & ofs ,
                       const MultiArrayView < 2 , ele_type > & weight
                     ) const
    {
      dtype result = dtype() ;
      for ( int i = 0 ; i < weight.shape(0) ; i++ )
      {
        result +=   weight [ Shape2 ( i , level ) ]
                  * _eval < level - 1 , dtype >() ( pdata , ofs , weight ) ;
      }
      return result ;
    }
  } ;

  /// at level 0 the recursion ends, now we finally apply the weights for axis 0
  /// to the window of coefficients. Note how ofs is passed in per reference. This looks
  /// wrong, but it's necessary: When, in the course of the recursion, the level 0
  /// routine is called again, it needs to access the next bunch of spline_order coefficients.
  /// Just incrementing the reference saves us incrementing higher up.
  
  template < class dtype >
  struct _eval < 0 , dtype >
  {
    dtype operator() ( const dtype* const & pdata ,
                       offset_iterator & ofs ,
                       const MultiArrayView < 2 , ele_type > & weight
                     ) const
    {
      dtype result = dtype() ;
      for ( int i = 0 ; i < weight.shape(0) ; i++ )
      {
        result += pdata [ *ofs ] * weight [ Shape2 ( i , 0 ) ] ;
        ++ofs ;
      }
      return result ;
    }
  } ;

  /// _eval_linear implements the specialization for n-linear (degree 1) b-splines.
  /// here, there is no gain to be had from working with precomputed per-axis weights,
  /// the weight generation is trivial. So the specialization given here is faster:
  
  template < int level , class dtype >
  struct _eval_linear
  {
    dtype operator() ( const dtype* const & pdata ,
                       offset_iterator & ofs ,
                       const nd_rc_type& tune
                     ) const
    {
      dtype result = dtype ( 1.0 - tune[level] )
                     * _eval_linear < level - 1 , dtype >() ( pdata , ofs , tune ) ;
      result += tune[level]
                * _eval_linear < level - 1 , dtype >() ( pdata , ofs , tune ) ;
      return result ;
    }
  } ;

  /// again, level 0 terminates the recursion
  
  template < class dtype >
  struct _eval_linear < 0 , dtype >
  {
    dtype operator() ( const dtype* const & pdata ,
                       offset_iterator & ofs ,
                       const nd_rc_type& tune
                     ) const
    {
      dtype result = dtype ( 1.0 - tune[0] ) * pdata [ *ofs ] ;
      ++ofs ;
      result += tune[0] * pdata [ *ofs ] ;
      ++ofs ;
      return result ;
    }
  } ;
    
  // next are evaluation routines. there are quite a few, since I've coded for operation
  // from different starting points and both for vectorized and nonvectorized operation.
  // Note how my code is, as ever, bottom-up, so that the highest-level access is below
  // the code it delegates to.

  // The code for vectorized and non-vectorized evaluation is often quite similar, in fact
  // so much so that I did an experimental implementation of a set of generic evaluators
  // which used the same code for vectorized and unverctorized operation. This implementation
  // had to use generic memory access functions which resulted in nD values being loaded
  // and stored in one go, rather than doing this channel-wise. I think that this made the
  // code slower (at least when compiled with clang++) and the necessary collateral code
  // to access memory was much more than what was saved in the evaluation code. So I
  // abandoned this approach for the time being.
  
  /// evaluation variant which takes an offset and a set of weights. This is the final delegate,
  /// calling the recursive _eval method. Having the weights passed in via a const MultiArrayView &
  /// allows calling code to provide their own weights, together with their shape, in one handy
  /// packet. And in the normal sequence of delegation inside class eval, the next routine 'up'
  /// can also package the weights nicely in a MultiArrayView. Note that select is now an ic_type,
  /// a single integral value representing an offset into the coefficient array.
  
  void eval ( const ic_type & select ,
              const MultiArrayView < 2 , ele_type > & weight ,
              value_type & result ) const
  {
    const value_type * const base = cf_base + select ;
    // offsets reflects the positions inside the subarray:
    offset_iterator ofs = offsets.begin() ;
    result = _eval<level,value_type>() ( base , ofs , weight ) ;
  }

  /// 'penultimate' delegate, taking a multidimensional index 'select' to the beginning of
  /// the coefficient window to process. Here the multidimensional index is translated
  /// into an offset from the coefficient array's base adress. Carrying the information as
  /// a reference to a multidimensional index all the way to this point does no harm, it's
  /// only a reference after all.

  void eval ( const nd_ic_type & select ,
              const MultiArrayView < 2 , ele_type > & weight ,
              value_type & result ) const
  {
    eval ( sum ( select * cf_stride ) , weight , result ) ;
  }

  /// evaluation variant taking the components of a split coordinate, first the shape_type
  /// representing the origin of the coefficient window to process, second the fractional parts
  /// of the coordinate which are needed to calculate the weights to apply to the coefficient
  /// window. Note that the weights aren't as many values as the window has coefficients, but
  /// only spline_order weights per dimension; the 'final' weights would be the outer product of the
  /// sets of weights for each dimension, which we don't explicitly use here. Also note that
  /// we limit the code here to have the same number of weights for each axis.
  /// Here we have the code for the specializations affected by the template argument
  /// 'specialize' (TODO silly name) which provide specialized code for degree 0 and 1
  /// b-splines (aka nearest-neighbour and n-linear interpolation)
  
  void eval ( const nd_ic_type& select ,
              const nd_rc_type& tune ,
              value_type & result ,
              std::integral_constant < int , 0 >
            ) const
  {
    // nearest neighbour. simply pick the coefficient
    result = cf_base [ sum ( select * cf_stride ) ] ;
  }
  
  void eval ( const nd_ic_type& select ,
              const nd_rc_type& tune ,
              value_type & result ,
              std::integral_constant < int , 1 >
            ) const
  {
    // linear interpolation. use specialized _eval_linear object
    const value_type * base =   cf_base
                              + sum ( select * cf_stride ) ;
    // offsets reflects the positions inside the subarray
    offset_iterator ofs = offsets.begin() ;
    result = _eval_linear<level,value_type>() ( base , ofs , tune ) ;
  }
  
  template < int arbitrary_spline_degree >
  void eval ( const nd_ic_type& select ,
              const nd_rc_type& tune ,
              value_type & result ,
              std::integral_constant < int , arbitrary_spline_degree >
            ) const
  {
    // general case. this works for degree 0 and 1 as well, but is less efficient

    MultiArray < 2 , ele_type > weight ( Shape2 ( spline_order , dimension ) ) ;
    
    // now we call obtain_weights, which will fill in 'weight' with weights for the
    // given set of fractional coordinate parts in 'tune'

    obtain_weights ( weight , tune ) ;
    eval ( select , weight , result ) ;   // delegate
  }

  /// this eval variant take an unsplit coordinate and splits it into an integral
  /// and a remainder part, using split(). The resulting 'split' coordinate is fed
  /// to the routine above.
  
  void eval ( const nd_rc_type& c , // unsplit coordinate
              value_type & result ) const
  {
    nd_ic_type select ;
    nd_rc_type tune ;
    
    // split coordinate into integral and remainder parts
    
    split ( c , select , tune ) ;
    
    // we now have the split coordinate in 'select'
    // and 'tune', and we delegate to the next level:

    eval ( select , tune , result , specialize() ) ;
  }
  
  /// variant taking a plain rc_type for a coordinate. only for 1D splines!
  /// This is to avoid hard-to-track errors which might ensue from allowing
  /// broadcasting of single rc_types to nd_rc_types for D>1. We convert the
  /// single coordinate (of rc_type) to an aggregate of 1 to fit the signature
  /// of the nD code above, which will then be optimized away again by the
  /// compiler.
  
  void eval ( const rc_type& c , // single 1D coordinate
              value_type & result ) const
  {
    static_assert ( dimension == 1 ,
                    "evaluation at a single real coordinate is only allowed for 1D splines" ) ;
    nd_rc_type cc ( c ) ;
    eval ( cc , result ) ;
  }

  /// variant of the above routine returning the result instead of depositing it
  /// in a reference
  
  value_type eval ( const rc_type& c ) const
  {
    static_assert ( dimension == 1 ,
                    "evaluation at a single real coordinate is only allowed for 1D splines" ) ;
    nd_rc_type cc ( c ) ;
    value_type result ;
    eval ( cc , result ) ;
    return result ;
  }

  /// alternative implementation of the last part of evaluation. Here, we calculate the
  /// outer product of the component vectors in 'weight' and use flat_eval to obtain the
  /// weighted sum over the coefficient window.

  /// evaluation using the outer product of the weight component vectors. This is simple,
  /// no need for recursion here, it's simply the sum of all values in the window, weighted
  /// with their corresponding weights. note that there are no checks; if there aren't enough
  /// weights the code may crash. Note that this code is formulated as a template and can be
  /// used both for vectorized and unvectorized operation.

  template < class dtype , class weight_type >
  void flat_eval ( const dtype * const & pdata ,
                   const weight_type * const pweight ,
                   value_type & result ) const
  {
    result = pweight[0] * pdata[offsets[0]] ;
    for ( int i = 1 ; i < window_size ; i++ )
      result += pweight[i] * pdata[offsets[i]] ;
  }

  /// calculation of the outer product of the component vectors of the weights:

  template < int _level , class dtype >
  struct outer_product
  {
    void operator() ( dtype * & target ,
                      const MultiArrayView < 2 , dtype > & weight ,
                      dtype factor
                    ) const
    {
      if ( _level == level ) // both are template args, this conditional has no runtime cost
      {
        // if _level == level this is the outermost call, and factor is certainly
        // 1.0. hence we can omit it:
        for ( int i = 0 ; i < weight.shape(0) ; i++ )
        {
          outer_product < _level - 1 , dtype >() ( target ,
                                                    weight ,
                                                    weight [ Shape2 ( i , level ) ] ) ;
        }
      }
      else
      {
        // otherwise we need it.
        for ( int i = 0 ; i < weight.shape(0) ; i++ )
        {
          outer_product < _level - 1 , dtype >() ( target ,
                                                    weight ,
                                                    weight [ Shape2 ( i , level ) ] * factor ) ;
        }
      }
    }
  } ;
  
  template < class dtype >
  struct outer_product < 0 , dtype >
  {
    void operator() ( dtype * & target ,
                      const MultiArrayView < 2 , dtype > & weight ,
                      dtype factor
                    ) const
    {
     if ( level == 0 )
      {
        // if level == 0, this is a 1D scenario, and all that is left is copying the weights
        // verbatim to target. TODO: for this special case (only occuring in 1D)
        // we could avoid this copy and use the data from 'weight' directly
        for ( int i = 0 ; i < weight.shape(0) ; i++ )
        {
          *target = weight [ Shape2 ( i , 0 ) ] ;
          ++target ;
        }
      }
      else
      {
        for ( int i = 0 ; i < weight.shape(0) ; i++ )
        {
          *target = weight [ Shape2 ( i , 0 ) ] * factor ;
          ++target ;
        }
      }
    }
  } ;

  /// evaluation variant which first calculates the final weights and then applies them
  /// using a simple summation loop. currently unused, since it is slower, but it can be
  /// used instead of eval() with the same signature.
  
  void flat_eval ( const ic_type & select ,
                   const MultiArrayView < 2 , ele_type > & weight ,
                   value_type & result ) const
  {
    // get a pointer to the coefficient window's beginning
    const value_type * base = cf_base + select ;
    // prepare space for the multiplied-out weights
    ele_type flat_weight [ window_size ] ;
    // we need an instance of a pointer to these weights here, passing it in
    // per reference to be manipulated by the code it's passed to
    ele_type * iter = flat_weight ;
    // now we multiply out the weights using outer_product()
    outer_product < level , ele_type >() ( iter , weight , 1.0 ) ;
    // finally we delegate to flat_eval above
    flat_eval ( base , flat_weight , result ) ;
  }

#ifdef USE_VC

  /// vectorized evaluation code. This works the same way as the unvectorized code, with the
  /// only difference being the innner loop over the channels, which is necessary because
  /// in the vector code we can't code for vectors of, say, pixels, but only for vectors of
  /// elementary types, like float.
  /// to operate with vsize values synchronously, we need a bit more indexing than in the
  /// non-vectorized version. the second parameter, origin, constitutes a gather operand
  /// which, applied to the base adress, handles a set of windows to be processed in parallel.
  /// if the gather operation is repeated with offsetted base addresses, the result vector is
  /// built in the same way as the single result value in the unvectorized code above.
  /// note that the vectorized routine couldn't function like this if it were to
  /// evaluate unbraced splines: it relies on the bracing and can't do without it, because
  /// it works with a fixed sequence of offsets, whereas the evaluation of an unbraced spline
  /// could use a different offset sequence for values affected by the boundary condition.

  typedef typename base_type::out_ele_v ele_v ;
  typedef typename base_type::out_v mc_ele_v ;
  typedef typename vector_traits < nd_ic_type , vsize > :: type nd_ic_v ;
  typedef typename vector_traits < nd_rc_type , vsize > :: type nd_rc_v ;
  
  template < class dtype , int level >
  struct _v_eval
  {
    dtype operator() ( const ele_type * const & ebase ,  ///< base adresses of components
                       const ic_v& origin ,              ///< offsets to evaluation window origins
                       offset_iterator & ofs ,           ///< offsets to coefficients inside this window
                       const MultiArrayView < 2 , ele_v > & weight ) const ///< weights to apply
    {
      dtype sum = dtype() ; ///< to accumulate the result
      dtype subsum ;        ///< to pick up the result of the recursive call

      for ( int i = 0 ; i < weight.shape ( 0 ) ; i++ )
      {
        subsum = _v_eval < dtype , level - 1 >() ( ebase , origin , ofs , weight );
        for ( int ch = 0 ; ch < channels ; ch++ )
        {
          sum[ch] += weight [ Shape2 ( i , level ) ] * subsum[ch] ;
        }
      }
      return sum ;
    }  
  } ;

  /// the level 0 routine terminates the recursion
  
  template < class dtype >
  struct _v_eval < dtype , 0 >
  {
    dtype operator() ( const ele_type * const & ebase ,  ///< base adresses of components
                       const ic_v& origin ,              ///< offsets to evaluation window origins
                       offset_iterator & ofs ,           ///< offsets to coefficients in this window
                       const MultiArrayView < 2 , ele_v > & weight ) const ///< weights to apply
    {
      dtype sum = dtype() ;

      for ( int i = 0 ; i < weight.shape ( 0 ) ; i++ )
      {
        for ( int ch = 0 ; ch < channels ; ch++ )
        {
          sum[ch] += weight [ Shape2 ( i , 0 ) ] * ele_v ( ebase + ch , origin + *ofs ) ;
        }
        ++ofs ;
      }
      return sum ;
    }  
  } ;

  // for linear (degree=1) interpolation we use a specialized routine
  // which is slightly faster, because it directly uses the 'tune' values
  // instead of building an array of weights first and passing that in.

  template < class dtype , int level >
  struct _v_eval_linear
  {
    dtype operator() ( const ele_type * const & ebase ,  ///< base adresses of components
                       const ic_v& origin ,        // offsets to evaluation window origins
                       offset_iterator & ofs ,     // offsets to coefficients inside this window
                       const in_v& tune ) const       // weights to apply
    {
      dtype sum ;    ///< to accumulate the result
      dtype subsum ;

      sum = _v_eval_linear < dtype , level - 1 >() ( ebase , origin , ofs , tune ) ;
      for ( int ch = 0 ; ch < channels ; ch++ )
      {
        sum[ch] *= ( rc_type ( 1.0 ) - tune [ level ] ) ;
      }
      
      subsum = _v_eval_linear < dtype , level - 1 >() ( ebase , origin , ofs , tune );
      for ( int ch = 0 ; ch < channels ; ch++ )
      {
        sum[ch] += ( tune [ level ] ) * subsum[ch] ;
      }
      
      return sum ;
    }  
  } ;

  /// the level 0 routine terminates the recursion
  
  template < class dtype >
  struct _v_eval_linear < dtype , 0 >
  {
    dtype operator() ( const ele_type * const & ebase ,  ///< base adresses of components
                       const ic_v& origin ,              // offsets to evaluation window origins
                       offset_iterator & ofs ,           // offsets to coefficients in this window
                       const in_v& tune ) const       // weights to apply
    {
      dtype sum ;
      auto o1 = *ofs ;
      ++ofs ;
      auto o2 = *ofs ;
      ++ofs ;
      
      for ( int ch = 0 ; ch < channels ; ch++ )
      {
        sum[ch] = ( rc_type ( 1.0 ) - tune [ 0 ] )
                  * ele_v ( ebase + ch , origin + o1 ) ;
                  
        sum[ch] += tune [ 0 ]
                   * ele_v ( ebase + ch , origin + o2 ) ; 
      }
      
      
      return sum ;
    }  
  } ;

  // vectorized variants of the evaluation routines:
  
  /// this is the vectorized version of the final delegate, calling _v_eval. The first
  /// argument is a vector of offsets to windows of coefficients, the vectorized equivalent
  /// of the single offset in the unvectorized routine.
  /// The second argument is a 2D array of vecorized weights, the third a reference to
  /// the result.
  
  void eval ( const ic_v& select ,  // offsets to lower corners of the subarrays
              const MultiArrayView < 2 , ele_v > & weight , // vectorized weights
              out_v & result ) const
  {
    // we need an instance of this iterator because it's passed into _v_eval by reference
    // and manipulated by the code there:
    
    offset_iterator ofs = eoffsets.begin() ;
    
    // now we can call the recursive _v_eval routine yielding the result
    
    result = _v_eval < out_v , level >() ( cf_ebase , select , ofs , weight ) ;
  }

  /// cdeval starts from a set of offsets to coefficient windows, so here
  /// the nD integral indices to the coefficient windows have already been 'condensed'
  /// into 1D offsets into the coefficient array's memory.
  /// KFJ 2017-08-20 I'd like tto change the code so that I can directly operate
  /// on 1D data instead of having to use TinyVector<T,1>. For this case, this
  /// eval overload and the next would have the same signature, so I changed the
  /// name to cdeval for the time being. cdeval might be factored into the next eval
  /// since I think it's not really necessary as a separate routine.
  /// Here we have the specializations affected by the template argument 'specialize'
  /// which activates more efficient code for degree 0 (nearest neighbour) and degree 1
  /// (linear interpolation) splines. I draw the line here; one might add further
  /// specializations, but from degree 2 onwards the weights are reused several times
  /// so looking them up in a small table (as the general-purpose code for unspecialized
  /// operation does) should be more efficient (TODO test).
  
  /// we have three variants, depending on 'specialize'. first is the specialization
  /// for nearest-neighbour interpolation, which doesn't delegate further, since the
  /// result can be obtained directly by gathering from the coefficients:
  
  void eval ( const ic_v& select ,  // offsets to coefficient windows
              const in_v& tune ,    // fractional parts of the coordinates
              out_v & result ,      // target
              std::integral_constant < int , 0 > ) const
  {
    // nearest neighbour. here, no weights are needed and we can instantly
    // gather the data from the coefficient array.

    for ( int ch = 0 ; ch < channels ; ch++ )
      result[ch].gather ( cf_ebase + ch , select ) ;
  }
  
  /// linear interpolation. this uses a specialized _v_eval object which
  /// directly processes 'tune' instead of creating an array of weights first.

  void eval ( const ic_v& select ,  // offsets to coefficient windows
              const in_v& tune ,    // fractional parts of the coordinates
              out_v & result ,      // target
              std::integral_constant < int , 1 > ) const
  {
    offset_iterator ofs = eoffsets.begin() ;
  
    result = _v_eval_linear < out_v , level >()
              ( cf_ebase , select , ofs , tune ) ;
   }
  
  /// finally, the general uniform b-spline evaluation.
  /// Passing any number apart from 0 or 1 as 'specialize' template argument results
  /// in the use of this general-purpose code, which can also handle degree 0 and 1
  /// splines, albeit less efficiently.

  template < int arbitrary_spline_degree >
  void eval ( const ic_v& select ,  // offsets to coefficient windows
              const in_v& tune ,    // fractional parts of the coordinates
              out_v & result ,      // target
              std::integral_constant < int , arbitrary_spline_degree > ) const
  {
    MultiArray < 2 , ele_v > weight ( Shape2 ( spline_order , dimension ) ) ;

    // obtain_weights is the same code as for unvectorized operation, the arguments
    // suffice to pick the right template arguments
  
    obtain_weights ( weight , tune ) ;
    
    // having obtained the weights, we delegate to the final delegate.
    
    eval ( select , weight , result ) ;
  }

  /// here we transform incoming nD coordinates into offsets into the coefficient
  /// array's memory. In my experiments I found that switching from nD indices
  /// to offsets is best done sooner rather than later, even though one  might suspect
  /// that simply passing on the reference to the nD index and converting it 'further down'
  /// shouldn't make much difference. Probably it's easier for the optimizer to
  /// see the conversion closer to the emergence of the nD index from the nD real
  /// coordinate coming in.
  /// note that we use both the strides and offsets appropriate for an expanded array,
  /// and component_base has pointers to the elementary type.

  void eval ( const nd_ic_v & select , // nD coordinates to coefficient windows
              const nd_rc_v & tune ,   // fractional parts of the coordinates
              out_v & result ) const
  {
    // condense the nD index into an offset
    ic_v origin = select[0] * ic_type ( cf_estride [ 0 ] ) ;
    for ( int d = 1 ; d < dimension ; d++ )
      origin += select[d] * ic_type ( cf_estride [ d ] ) ;
    
    // pass on to overload taking the offset, dispatching on 'specialize'
    eval ( origin , tune , result , specialize() ) ;
  }

  /// This variant of eval() works directly on vector data (of unsplit coordinates)
  /// This burdens the calling code with (de)interleaving the data. But often the calling
  /// code performs a traversal of a large body of data and is therefore in a better position
  /// to perform the (de)interleaving e.g. by a gather/scatter operation, or already receives
  /// the data in simdized form. This latter case is actually quite common, because 'real'
  /// applications will rarely use class evaluator directly. Instead, the evaluator will usually
  /// be used as some 'inner' component of another functor, which is precisely why class
  /// evaluator is implemented as a 'pure' functor, containing only state fixed at construction.
  /// So whatever deinterleaving and preprocessing and postprocessing and reinterleaving
  /// may be performed 'outside' is irrelevant here where we receive SIMD data only.
  /// In this routine, we make the move from incoming real coordinates to separate
  /// nD integral indices and fractional parts.
  
  void eval ( const in_v & input ,    // number of dimensions * coordinate vectors
              out_v & result )  const // number of channels * value vectors
  {
    nd_ic_v select ;
    in_v tune ;

    // split the coordinate into integral and an remainder part
    
    split ( input , select , tune ) ;
    
    // delegate to eval with split coordinates

    eval ( select , tune , result ) ;
  }

  /// To make the evaluator more versatile, we provide overloads taking simdized elementary
  /// type as well, where unary_functor would expect TinyVectors < T , 1 >.
  
  void eval ( const in_v & input ,        // number of dimensions * coordinate vectors
              out_ele_v & result )  const // single value vector
  {
    static_assert ( dim_out == 1 ,
                    "this evaluation routine is for single-channel data only" ) ;
    out_v helper ;
    eval ( input , helper ) ;
    result = helper[0] ;
  }

  /// ditto, for 1D coordinates

  void eval ( const in_ele_v & input , // single coordinate vector
              out_v & result ) const   // simdized mD output
  {
    static_assert ( dim_in == 1 ,
                    "this evaluation routine is for 1D coordinates only" ) ;
    in_v helper ;
    helper[0] = input ;
    eval ( helper , result ) ;
  }

  void eval ( const in_ele_v & input ,    // single coordinate vector
              out_ele_v & result )  const // single value vector
  {
    static_assert ( dim_in == 1 && dim_out == 1  ,
                    "this evaluation routine is for single-channel data and 1D coordinates only") ;
    in_v in_helper ;
    in_helper[0] = input ;
    out_v out_helper ;
    eval ( in_helper , out_helper ) ;
    result = out_helper[0] ;
  }

#endif // USE_VC

  ~evaluator_policy()
  {
    // we don't want a memory leak!
    for ( int d = 0 ; d < dimension ; d++ )
    {
      if ( fweight[d] != &wfd0 )
        delete fweight[d] ;
    }
  }
} ;

/// the definition for vspline::evaluator incorporates the policy class
/// into a vspline::unary_functor. this is quite a mouthful, so we use a
/// using declaration:

template < typename _coordinate_type , // nD real coordinate
           typename _value_type ,      // type of coefficient/result
#ifdef USE_VC
           // nr. of vector elements with sensible default
           int _vsize = vspline::vector_traits < _value_type > :: size ,
#else
           int _vsize = 1 ,
#endif
           typename specialize = std::integral_constant<int,-1>
         >
using evaluator = unary_functor < _coordinate_type ,
                                  _value_type ,
                                  _vsize ,
                                  evaluator_policy ,
                                  specialize > ;

} ; // end of namespace vspline

#endif // VSPLINE_EVAL_H
