/************************************************************************/
/*                                                                      */
/*    vspline - a set of generic tools for creation and evaluation      */
/*              of uniform b-splines                                    */
/*                                                                      */
/*            Copyright 2015 - 2017 by Kay F. Jahnke                    */
/*                                                                      */
/*    The git repository for this software is at                        */
/*                                                                      */
/*    https://bitbucket.org/kfj/vspline                                 */
/*                                                                      */
/*    Please direct questions, bug reports, and contributions to        */
/*                                                                      */
/*    kfjahnke+vspline@gmail.com                                        */
/*                                                                      */
/*    Permission is hereby granted, free of charge, to any person       */
/*    obtaining a copy of this software and associated documentation    */
/*    files (the "Software"), to deal in the Software without           */
/*    restriction, including without limitation the rights to use,      */
/*    copy, modify, merge, publish, distribute, sublicense, and/or      */
/*    sell copies of the Software, and to permit persons to whom the    */
/*    Software is furnished to do so, subject to the following          */
/*    conditions:                                                       */
/*                                                                      */
/*    The above copyright notice and this permission notice shall be    */
/*    included in all copies or substantial portions of the             */
/*    Software.                                                         */
/*                                                                      */
/*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND    */
/*    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES   */
/*    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND          */
/*    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT       */
/*    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,      */
/*    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING      */
/*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR     */
/*    OTHER DEALINGS IN THE SOFTWARE.                                   */
/*                                                                      */
/************************************************************************/

/*! \file map.h

    \brief code to handle out-of-bounds coordinates.
    
    Incoming coordinates may not be inside the range which can be evaluated
    by a functor. There is no one correct way of dealing with out-of-bounds
    coordinates, so we provide a few common ways of doing it.
    
    If the 'standard' gate types don't suffice, users can easily write their
    own handling code by providing gate objects or mapper objects derived
    from vspline::unary_functor and chaining them with their functors,
    the classes provided here can serve as templates.
    
    The basic type handling the operation is 'gate_type', which 'treats'
    a single value or single simdized type. For nD coordinates, we use a
    set of these gate_type objects, one for each component; each one may be
    of a distinct type specific to the axis the component belongs to.
    
    Application of the gates is via a 'mapper' object, which contains
    the gate_types and applies them to the components in turn.
    
    The final mapper object is a functor which converts an arbitrary incoming
    coordinate into a 'treated' coordinate (or, for REJECT mode, may throw an
    out_of_bounds exception). If the treatment succeeds to produce a suitable
    in-range coordinate, accessing the inner functor with it is safe.
    
    gate_type objects are derived from vspline::unary_functor, so they fit in
    well with other code in vspline and can easily be wrapped in other
    unary_functor objects, or used stand-alone. class mapper is also derived
    from unary_functor. With all functors concerned derived from unary_functor,
    we can use class vspline::chain (in unary_functor.h) to combine the mapper
    and the inner functor.

*/

#ifndef VSPLINE_MAP_H
#define VSPLINE_MAP_H

#include <vspline/unary_functor.h>
#include <vigra/tinyvector.hxx>

namespace vspline
{
/// gate with reject_policy throws vspline::out_of_bounds for invalid coordinates

template < typename _in_type ,
           typename _out_type ,
           int _vsize >
struct reject_policy
: public vspline::uf_types < _in_type , _out_type , _vsize >
{
  typedef _in_type rc_type ;
  typedef vspline::uf_types < rc_type , rc_type , _vsize > base_type ;
  using_unary_functor_types(base_type) ;

  const rc_type lower ;
  const rc_type upper ;
  
  reject_policy ( rc_type _lower ,
                rc_type _upper )
  : lower ( _lower ) ,
    upper ( _upper )
  { } ;

  void eval ( const in_type & c ,
                    out_type & result ) const
  {
    if ( c < lower || c > upper )
      throw vspline::out_of_bounds() ;
    result = c ;
  }
  
#ifdef USE_VC

  void eval ( const in_ele_v & c ,
                    out_ele_v & result ) const
  {
    if ( any_of ( ( c < lower ) | ( c > upper ) ) )
      throw vspline::out_of_bounds() ;
    result = c ;
  }

#endif

} ;

/// gate with limit_policy clamps out-of-bounds values

template < typename _in_type ,
           typename _out_type ,
           int _vsize >
struct limit_policy
: public vspline::uf_types < _in_type , _out_type , _vsize >
{
  typedef _in_type rc_type ;
  typedef vspline::uf_types < rc_type , rc_type , _vsize > base_type ;
  using_unary_functor_types(base_type) ;
  
  const rc_type lower ;
  const rc_type upper ;
  
  limit_policy ( rc_type _lower ,
                rc_type _upper )
  : lower ( _lower ) ,
    upper ( _upper )
  { } ;

  void eval ( const in_type & c ,
                    out_type & result ) const
  {
    if ( c < lower )
      result = lower ;
    else if ( c > upper )
      result = upper ;
    else
      result = c ;
  }
  
#ifdef USE_VC

  void eval ( const in_ele_v & c ,
                    out_ele_v & result ) const
  {
    out_ele_v cc = c ;
    cc ( cc < lower ) = lower ;
    cc ( cc > upper ) = upper ;
    result = cc ;
  }

#endif

} ;

/// constant gate assigns 'fix' to out-of-bounds coordinates; by default
/// it will return 0 for out-of-bounds input, but the value can be chosen
/// by passing a third argument to the constructor.

template < typename _in_type ,
           typename _out_type ,
           int _vsize >
struct constant_policy
: public vspline::uf_types < _in_type , _out_type , _vsize >
{
  typedef _in_type rc_type ;
  typedef vspline::uf_types < rc_type , rc_type , _vsize > base_type ;
  using_unary_functor_types(base_type) ;
  
  const rc_type lower ;
  const rc_type upper ;
  const rc_type fix ;
  
  constant_policy ( rc_type _lower ,
                rc_type _upper ,
                rc_type _fix = rc_type(0)
            )
  : lower ( _lower ) ,
    upper ( _upper ) ,
    fix ( _fix )
  { } ;

  void eval ( const in_type & c ,
                    out_type & result ) const
  {
    if ( c < lower )
      result = fix ;
    else if ( c > upper )
      result = fix ;
    else
      result = c ;
  }
  
#ifdef USE_VC

  void eval ( const in_ele_v & c ,
                    out_ele_v & result ) const
  {
    out_ele_v cc = c ;
    cc ( cc < lower ) = fix ;
    cc ( cc > upper ) = fix ;
    result = cc ;
  }

#endif

} ;

#ifdef USE_VC

template <typename rc_v>
rc_v v_fmod ( const rc_v& lhs ,
              const typename rc_v::EntryType rhs )
{
  typedef typename vector_traits < int , rc_v::Size > :: ele_v ic_v ;

  ic_v iv ( lhs / rhs ) ;
  return lhs - rhs * rc_v ( iv ) ;
}

#endif

/// gate with mirror_policy 'folds' coordinates into the range. It places the
/// result value so that mirroring it on both lower and upper (which
/// produces an infinite number of mirror images) will produce one
/// mirror image coinciding with the input.
/// When using this gate type with splines with MIRROR boundary conditions,
/// if the shape of the core for the axis in question is M, _lower would be
/// passed 0 and _upper M-1.
/// For splines with REFLECT boundary conditions, we'd pass -0.5 to
/// _lower and M-0.5 to upper, since here we mirror 'between bounds'
/// and the defined range is wider.

template < typename _in_type ,
           typename _out_type ,
           int _vsize >
struct mirror_policy
: public vspline::uf_types < _in_type , _out_type , _vsize >
{
  typedef _in_type rc_type ;
  typedef vspline::uf_types < rc_type , rc_type , _vsize > base_type ;
  using_unary_functor_types(base_type) ;
  
  const rc_type lower ;
  const rc_type upper ;
  
  mirror_policy ( rc_type _lower ,
                rc_type _upper )
  : lower ( _lower ) ,
    upper ( _upper )
  { } ;

  void eval ( const in_type & c ,
                    out_type & result ) const
  {
    in_type cc = c ;
    if ( cc < lower )
      cc = 2 * lower - cc ;
    if ( cc > upper )
    {
      cc -= lower ;
      cc = std::fmod ( cc , 2 * ( upper - lower ) ) ;
      cc += lower ;
      if ( cc > upper )
        cc = 2 * upper - cc ;
    }
    result = cc ;
  }
  
#ifdef USE_VC

/// vectorized fmod function

  void eval ( const in_ele_v & c ,
                    out_ele_v & result ) const
  {
    out_ele_v cc ;
    
    cc = c - lower ;
    auto w = upper - lower ;

    cc = abs ( cc ) ;               // left mirror, v is now >= 0

    if ( any_of ( cc > w ) )
    {
      cc = v_fmod ( cc , 2 * w ) ;  // map to one full period
      cc -= w ;                     // center
      cc = abs ( cc ) ;             // map to half period
      cc = w - cc ;                 // flip
    }
    
    result = cc + lower ;
  }

#endif

} ;

/// the periodic mapping also folds the incoming value into the allowed range.
/// The resulting value will be ( N * period ) from the input value and inside
/// the range, period being upper - lower.
/// For splines done with PERIODIC boundary conditions, if the shape of
/// the core for this axis is M, we'd pass 0 to _lower and M to _upper.

template < typename _in_type ,
           typename _out_type ,
           int _vsize >
struct periodic_policy
: public vspline::uf_types < _in_type , _out_type , _vsize >
{
  typedef _in_type rc_type ;
  typedef vspline::uf_types < rc_type , rc_type , _vsize > base_type ;
  using_unary_functor_types(base_type) ;
  
  const rc_type lower ;
  const rc_type upper ;
  
  periodic_policy ( rc_type _lower ,
                rc_type _upper )
  : lower ( _lower ) ,
    upper ( _upper )
  { } ;

  void eval ( const in_type & c ,
                    out_type & result ) const
  {
    in_type cc = c - lower ;
    auto w = upper - lower ;
    
    if ( cc < 0 )
      cc = w + fmod ( cc , w ) ;
    else if ( cc >= w )
      cc = fmod ( cc , w ) ;
    result = cc + lower ;
  }
  
#ifdef USE_VC

  void eval ( const in_ele_v & c ,
                    out_ele_v & result ) const
  {
    out_ele_v cc ;
    
    cc = c - lower ;
    auto w = upper - lower ;

    if ( any_of ( ( cc < 0 ) | ( cc >= w ) ) )
    {
      cc = v_fmod ( cc , w ) ;
      cc ( cc < 0 ) += w ;
    }
    
    result = cc + lower ;
  }

#endif

} ;

/// using declaration for the gate type. This declaration incoporates
/// a mapping policy into a vspline::unary_functor:

template < typename coordinate_type ,
           int _vsize ,
           template < class ,
                      class ,
                      int > class gate_policy >
using gate_type =
  vspline::unary_functor < coordinate_type ,
                           coordinate_type ,
                           _vsize ,
                            gate_policy > ;

template < typename coordinate_type ,
           int _vsize = 1 >
using reject_gate_type =
vspline::gate_type < coordinate_type ,        // coordinate type
                     _vsize ,                 // vector size
                     vspline::reject_policy // mapping policy
                   > ;

template < typename coordinate_type ,
           int _vsize = 1 >
using limit_gate_type =
vspline::gate_type < coordinate_type ,        // coordinate type
                     _vsize ,                 // vector size
                     vspline::limit_policy // mapping policy
                   > ;

template < typename coordinate_type ,
           int _vsize = 1 >
using constant_gate_type =
vspline::gate_type < coordinate_type ,        // coordinate type
                     _vsize ,                 // vector size
                     vspline::constant_policy // mapping policy
                   > ;

template < typename coordinate_type ,
           int _vsize = 1 >
using mirror_gate_type =
vspline::gate_type < coordinate_type ,        // coordinate type
                     _vsize ,                 // vector size
                     vspline::mirror_policy // mapping policy
                   > ;

template < typename coordinate_type ,
           int _vsize = 1 >
using periodic_gate_type =
vspline::gate_type < coordinate_type ,        // coordinate type
                     _vsize ,                 // vector size
                     vspline::periodic_policy // mapping policy
                   > ;

/// finally we define class mapper which is initialized with a set of
/// gate objects (of arbitrary type) which are applied to each component
/// of an incoming nD coordinate in turn.
/// The trickery with the variadic template argument list is necessary,
/// because we want to be able to combine arbitrary gate types, which
/// have distinct types to make the mapper as efficient as possible.
/// the only requirement for a gate type is that it has to provide the
/// necessary eval() functions. Typically a gate type would be a 1D
/// vspline::unary_functor, like the types above, since this guarantees
/// a suitable type, but this is not enforced.

// TODO: I'd like to be able to construct a mapper object from
// a vspline::bspline (which provides lower_limit() and upper_limit()
// for each axis)

template < typename _in_type ,
           typename _out_type ,
           int _vsize ,
           class ... gate_types >
struct mapper_policy
{
  enum { dimension = vigra::ExpandElementResult < _in_type > :: size } ;
  
  // we hold the 1D mappers in a tuple
  
  typedef std::tuple < gate_types... > mvec_type ;
  
  // mvec holds the 1D gate objects passed to the constructor
  
  const mvec_type mvec ;
  
  // the constructor receives gate objects

  mapper_policy ( gate_types ... args )
  : mvec ( args... )
  { } ;
  
  // to handle the application of the 1D gates, we use a recursive
  // helper type which applies the 1D gate for a specific axis and
  // then recurses to the next axis until axis 0 is reached.
  // We also pass 'dimension' as template argument, so we can specialize
  // for 1D operation (see below)

  template < int level , int dimension , typename nd_coordinate_type >
  struct _map
  { 
    void operator() ( const mvec_type & mvec ,
                      const nd_coordinate_type & in ,
                      nd_coordinate_type & out ) const
    {
      std::get<level>(mvec).eval ( in[level] , out[level] ) ;
      _map < level - 1 , dimension , nd_coordinate_type >() ( mvec , in , out ) ;
    }
  } ;
  
  // at level 0 the recursion ends
  
  template < int dimension , typename nd_coordinate_type >
  struct _map < 0 , dimension , nd_coordinate_type >
  { 
    void operator() ( const mvec_type & mvec ,
                      const nd_coordinate_type & in ,
                      nd_coordinate_type & out ) const
    {
      std::get<0>(mvec).eval ( in[0] , out[0] ) ;
    }
  } ;
  
  // here's the specialization for 1D operation

  template < typename coordinate_type >
  struct _map < 0 , 1 , coordinate_type >
  { 
    void operator() ( const mvec_type & mvec ,
                      const coordinate_type & in ,
                      coordinate_type & out ) const
    {
      std::get<0>(mvec).eval ( in , out ) ;
    }
  } ;

  // now we define eval for unvectorized and vectorized operation
  // by simply delegating to struct _map at the top level.

  template < class in_type , class out_type >
  void eval ( const in_type & in ,
                    out_type & out ) const
  {
    _map < dimension - 1 , dimension , in_type >() ( mvec , in , out ) ;
  }

} ;

/// this using declaration incorporates the types of the gates
/// into the mapper, which is a vspline::unary_functor.

template < typename _coordinate_type ,
           int _vsize ,
           class ... gate_types >
using mapper = vspline::unary_functor < _coordinate_type ,
                                        _coordinate_type ,
                                        _vsize ,
                                        mapper_policy ,
                                        gate_types ... > ;

} ; // namespace vspline

#endif // #ifndef VSPLINE_MAP_H
